class Resource:

    def links(self):
        """
        Returns a list of all links.
        """
        raise NotImplementedError('Subclasses must implement this method')

    def link(self, rel):
        """
        Return a Link with rel.
        """
        raise NotImplementedError('Subclasses must implement this method')



import dsl


class OpenSearchDescription(object):
    """
    OpenSearchDescription object wraps the OpenSearch logic
    """

    def __init__(self, element_tree):
        self.url_type = 'application/rss+xml'
        self.element_tree = element_tree

    def use(self, url_type):
        """
        Set the OpenSearch type
        """
        self.url_type = url_type
        return self

    def search(self, searchTerms, startPage):
        """
        Make a search with 'searchTerms'
        It will find the url_type URL that you have chosen
        """
        tag = '{http://a9.com/-/spec/opensearch/1.1/}Url'
        urls = self.element_tree.findall(tag)
        for url in urls:
            if url.get('type') == self.url_type:
                template = url.get('template')
                template = template.replace('{searchTerms}', searchTerms)
                template = template.replace('{startPage?}', str(startPage))
                return dsl.Dsl(template).accepts(self.url_type).get()



from response import LazyResponse
from parser import Parser
from threading import Thread
from multiprocessing import Pipe


class Request(object):
    """
    HTTP request.
    """

    def __init__(self, config):
        """
        Initialize an HTTP request instance for a given configuration.
        """
        self.config = config

    def __call__(self, **kwargs):
        """
        Perform the request

        The optional payload argument is sent to the server.
        """
        if (not self.config.is_async):
            return self._process_flow(kwargs)
        else:
            return self._process_async_flow(kwargs)

    def _process_flow(self, payload):
        """
        Put payload environment and start the chain.
        """
        env = {}
        if payload:
            env = {'payload': payload}

        procs = list(self.config.processors)
        return Parser(procs).follow(self.config, env)

    def _process_async_flow(self, payload):
        """
        Starts an async chain.
        """

        self.config.pipe, child_pipe = Pipe()

        def handle_async():
            if self.config.is_async and self.config.callback is None:
                self._process_flow(payload=payload)
            else:
                self.config.callback(self._process_flow(payload=payload), \
                                     *self.config.callback_args)

        self._start_new_thread(handle_async)

        return LazyResponse(child_pipe)

    def _start_new_thread(self, target):
        thread = Thread(target=target)
        thread.start()



from restfulie.resources import Resource
from restfulie.links import Links, Link

class XMLResource(Resource):
    """
    This resource is returned when a XML is unmarshalled.
    """

    def __init__ (self, element_tree):
        self.element_tree = element_tree
        self._links = Links(self._parse_links())
        self._enhance_element_tree()

    def _enhance_element_tree(self):
        """
        Enables access to XMLResources attributes with 'dot'.
        """
        setattr(self, "tag", self.element_tree.tag)

        for root_child in list(self.element_tree):
            if root_child.tag != 'link':
                if len(self.element_tree.findall(root_child.tag)) > 1:
                    setattr(self, root_child.tag, self.element_tree.findall(root_child.tag))
                elif len(list(root_child)) == 0:
                    setattr(self, root_child.tag, root_child.text)
                else:
                    setattr(self, root_child.tag, self.element_tree.find(root_child.tag))

        for element in self.element_tree.getiterator():
            for child in list(element):
                if len(element.findall(child.tag)) > 1:
                    setattr(element, child.tag, element.findall(child.tag))
                elif len(list(child)) == 0:
                    setattr(element, child.tag, child.text)
                else:
                    setattr(element, child.tag, element.find(child.tag))

    def _parse_links(self):
        """
        Find links in a ElementTree
        """
        links = []
        for element in self.element_tree.getiterator('link'):
            link = Link(href=element.attrib.get('href'),
                        rel=element.attrib.get('rel'),
                        content_type=element.attrib.get('type') or 'application/xml')

            links.append(link)

        return links

    def links(self):
        return self._links

    def link(self, rel):
        return self.links().get(rel)



from processor import RedirectProcessor, PayloadMarshallingProcessor, \
    ExecuteRequestProcessor, AuthenticationProcessor
from request import Request


class Dsl(object):
    """
    Configuration object for requests at a given URI.
    """

    HTTP_VERBS = ['delete', 'get', 'head', 'options', 'patch', 'post', 'put',
                  'trace']

    def __init__(self, uri):
        """
        Initialize the configuration for requests at the given URI.
        """
        self.credentials = None
        self.uri = uri
        self.processors = [AuthenticationProcessor(),
                           RedirectProcessor(),
                           PayloadMarshallingProcessor(),
                           ExecuteRequestProcessor(), ]
        self.headers = {'Content-Type': 'application/xml',
                        'Accept': 'application/xml'}
        self.is_async = False
        self.callback = None
        self.callback_args = ()

    def __getattr__(self, name):
        """
        Perform an HTTP request. This method supports calls to the following
        methods: delete, get, head, options, patch, post, put, trace

        Once the HTTP call is performed, a response is returned (unless the
        async method is used).
        """
        if (self._is_verb(name)):
            self.verb = name.upper()
            return Request(self)
        else:
            raise AttributeError(name)

    def _is_verb(self, name):
        """
        Checks if a string is a HTTP verb
        """
        return name in self.HTTP_VERBS

    def use(self, feature):
        """
        Register a feature at this configuration.
        """
        self.processors.insert(0, feature)
        return self

    def async(self, callback=None, args=()):
        """
        Use asynchronous calls. A HTTP call performed through this object will
        return immediately, giving None as response. Once the request is
        completed, the callback function is called and the response and the
        optional extra args defined in args are passed as parameters.
        """
        self.is_async = True
        self.callback = callback
        self.callback_args = args
        return self

    def as_(self, content_type):
        """
        Set up the Content-Type
        """
        if content_type:
            self.headers["Content-Type"] = content_type
        return self

    def accepts(self, content_type):
        """
        Configure the accepted response format.
        """
        self.headers['Accept'] = content_type
        return self

    def auth(self, username, password, method='simple'):
        """
        Authentication feature. It does simple HTTP auth
        """
        self.credentials = (username, password, method)
        return self



import dsl


class Link(object):
    """
    Link represents generic link. You can follow it.
    """

    def __init__(self, href, rel, content_type='application/xml'):
        self.href = href
        self.rel = rel
        self.content_type = content_type

    def follow(self):
        """
        Return a DSL object with the Content-Type
        set.
        """
        return dsl.Dsl(self.href).as_(self.content_type)


class Links:
    """
    Links a simple Link collection. There are some
    methods to put syntax sugar
    """

    def __init__(self, links):
        """
        Enable Links to access attributes with 'dot'
        """
        self.links = {}
        for link in links:
            self.links[link.rel] = link
            setattr(self, link.rel, link)

    def get(self, rel):
        """
        Checks if a Link exists. If exists returns the
        object, else returns None
        """
        return self.links.get(rel)

    def __len__(self):
        """
        The length of Links is the length of the links
        dictionary inside it
        """
        return len(self.links)



from restfulie.resources import Resource
from restfulie.links import Links, Link

class JsonResource(Resource):
    """
    This resource is returned when a JSON is unmarshalled.
    """

    def __init__(self, dict_):
        """
        JsonResource attributes can be accessed with 'dot'.
        """
        links = self._parse_links(dict_)
        self._links = Links(links)

        for key, value in dict_.items():
            if isinstance(value, (list, tuple)):
                d = [JsonResource(x) if isinstance(x, dict) else x for x in value]
                setattr(self, key, d)
            else:
                d = JsonResource(value) if isinstance(value, dict) else value
                setattr(self, key, d)

    def _find_dicts_in_dict(self, structure):
        """
        Get all dictionaries on a structure and returns a list of it.
        """
        dicts = []
        if isinstance(structure, dict):
            dicts.append(structure)
            for k, v in structure.items():
                dicts.extend(self._find_dicts_in_dict(v))
        return dicts

    def _parse_links(self, dict_):
        """
        Find links on JSON dictionary.
        """
        for d in self._find_dicts_in_dict(dict_):
            if 'link' in d:
                return [Link(href=link.get('href'), rel=link.get('rel'), content_type=link.get('type')) for link in d['link']]

        return []

    def links(self):
        return self._links

    def link(self, rel):
        return self.links().get(rel)



import json
from xml.etree import ElementTree
from opensearch import OpenSearchDescription
from resources.xml import XMLResource
from resources.json import JsonResource

class Converters(object):
    """
    Utility methods for converters.
    """

    types = {}

    @staticmethod
    def register(a_type, converter):
        """
        Register a converter for the given type.
        """
        Converters.types[a_type] = converter

    @staticmethod
    def marshaller_for(a_type):
        """
        Return a converter for the given type.
        """
        return Converters.types.get(a_type) or XmlConverter()


class JsonConverter(object):
    """
    Converts objects from and to JSON.
    """

    def marshal(self, content):
        """
        Produces a JSON representation of the given content.
        """
        return json.dumps(content)

    def unmarshal(self, json_content):
        """
        Produces an object for a given JSON content.
        """
        return JsonResource(json.loads(json_content))

class XmlConverter(object):
    """
    Converts objects from and to XML.
    """

    def marshal(self, content):
        """
        Produces a XML representation of the given content.
        """
        return ElementTree.tostring(self._dict_to_etree(content))

    def _dict_to_etree(self, content):
        """
        Receives a dictionary and converts to an ElementTree
        """
        tree = ElementTree.Element(content.keys()[0])
        self._dict_to_etree_rec(content[content.keys()[0]], tree)
        return tree

    def _dict_to_etree_rec(self, content, tree):
        """
        Auxiliar function of _dict_to_etree_rec
        """
        if type(content) == dict:
            for key, value in content.items():
                e = ElementTree.Element(key)
                self._dict_to_etree_rec(value, e)
                tree.append(e)
        else:
            tree.text = str(content)

    def unmarshal(self, content):
        """
        Produces an ElementTree object for a given XML content.
        """
        e = ElementTree.fromstring(content)
        return XMLResource(e)


class OpenSearchConverter(object):
    def marshal(self, content):
        return XmlConverter().marshal(content)

    def unmarshal(self, content):
        """
        Produces an OpenSearchDescription object from an
        OpenSearch XML
        """
        e_tree = ElementTree.fromstring(content)
        return OpenSearchDescription(e_tree)


class PlainConverter(object):
    def marshal(self, content):
        """
        Does nothing
        """
        return content

    def unmarshal(self, content):
        """
        Returns content without modification
        """
        return content

Converters.register('application/xml', XmlConverter())
Converters.register('text/xml', XmlConverter())
Converters.register('xml', XmlConverter())
Converters.register('text/plain', PlainConverter())
Converters.register('text/json', JsonConverter())
Converters.register('application/json', JsonConverter())
Converters.register('json', JsonConverter())
Converters.register('application/opensearchdescription+xml',
                    OpenSearchConverter())



from converters import Converters
import re

from links import Links, Link


class Response(object):
    """
    Handle and parse a HTTP response
    """

    def __init__(self, response):
        self.response = response
        self.headers = self.response[0]
        self.code = self.response[0]['status']
        self.body = self.response[1]

    def resource(self):
        """
        Returns the unmarshalled object of the response body
        """
        if 'content-type' in self.response[0]:
            contenttype = self.response[0]['content-type'].split(';')[0]
        else:
            contenttype = None

        converter = Converters.marshaller_for(contenttype)
        return converter.unmarshal(self.body)

    def links(self):
        """
        Returns the Links of the header
        """
        r = self._link_header_to_array()
        return Links(r)

    def link(self, rel):
        """
        Get a link with 'rel' from header
        """
        return self.links().get(rel)

    def _link_header_to_array(self):
        """
        Split links in headers and return a list of dicts
        """
        values = self.headers['link'].split(',')
        links = []
        for link_string in values:
            links.append(self._string_to_link(link_string))

        return links

    def _string_to_link(self, l):
        """
        Parses a link header string to a dictionary
        """
        uri = re.search('<([^>]*)', l) and re.search('<([^>]*)', l).group(1)
        rest = re.search('.*>(.*)', l) and re.search('.*>(.*)', l).group(1)
        rel = (re.search('rel=(.*)', rest) and
               re.search('rel="(.*)"', rest).group(1))
        tpe = (re.search('type=(.*)', rest) and
               re.search('type="(.*)"', rest).group(1))

        return Link(href=uri, rel=rel, content_type=tpe)


class LazyResponse(object):
    """
    Lazy response for async calls
    """

    def __init__(self, response_pipe):
        self._response_pipe = response_pipe

    def __getattr__(self, attr):
        if self._response_pipe is not None:
            self._response = self._response_pipe.recv()
            self._response_pipe = None
        return getattr(self._response, attr)



class Parser(object):
    """
    Executes processors ordered by the list
    """

    def __init__(self, processors):
        self.processors = processors

    def follow(self, request, env={}):
        processor = self.processors.pop(0)
        result = processor.execute(self, request, env)
        return result



#!/usr/bin/python

from setuptools import setup, find_packages

setup(name='restfulie',
      version='0.9.2',
      description='Writing hypermedia aware resource based clients and servers',
      author=' ',
      author_email=' ',
      url='http://restfulie.caelumobjects.com/',
      long_description='CRUD through HTTP is a good step forward to using resources and becoming RESTful, another step further is to make use of hypermedia aware resources and Restfulie allows you to do it in Python.',
      download_url='https://github.com/caelum/restfulie-py',
      keywords='restfulie rest http hypermedia',
      classifiers=[
          "Development Status :: 4 - Beta",
          "Environment :: Web Environment",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Apache Software License",
          "Operating System :: MacOS :: MacOS X",
          "Operating System :: Microsoft :: Windows",
          "Operating System :: POSIX",
          "Programming Language :: Python",
      ],
      test_suite = "nose.collector",
      install_requires= ['httplib2>=0.6.0'],
      packages=find_packages(),
      include_package_data=True,
      )





from base64 import encodestring
import httplib2
from response import Response
from converters import Converters
import restfulie


class RequestProcessor(object):
    def execute(self, chain, request, env={}):
        raise NotImplementedError('Subclasses must implement this method')


class AuthenticationProcessor(RequestProcessor):
    """
    Processor responsible for making HTTP simple auth
    """
    def execute(self, chain, request, env={}):
        if request.credentials is not None:
            encoded_credentials = self._encode_credentials(request.credentials)
            request.headers['authorization'] = "Basic %s" % encoded_credentials
        return chain.follow(request, env)

    def _encode_credentials(self, credentials):
        username = credentials[0]
        password = credentials[1]
        method = credentials[2]
        if (method == 'simple'):
            return encodestring("%s:%s" % (username, password))[:-1]


class ExecuteRequestProcessor(RequestProcessor):
    """
    Processor responsible for getting the body from environment and
    making a request with it.
    """
    def __init__(self):
        self.http = httplib2.Http()

    def execute(self, chain, request, env={}):
        if "body" in env:
            response = self.http.request(request.uri, request.verb,
                                         env.get("body"), request.headers)
        else:
            response = self.http.request(request.uri, request.verb,
                                         headers=request.headers)

        resource = Response(response)
        if request.is_async:
            request.pipe.send(resource)

        return resource


class PayloadMarshallingProcessor(RequestProcessor):
    """
    Processor responsible for marshalling the payload in environment.
    """
    def execute(self, chain, request, env={}):
        if "payload" in env:
            content_type = request.headers.get("Content-Type")
            marshaller = Converters.marshaller_for(content_type)
            env["body"] = marshaller.marshal(env["payload"])
            del(env["payload"])

        return chain.follow(request, env)


class RedirectProcessor(RequestProcessor):
    """
    A processor responsible for redirecting a client to another URI when the
    server returns the location header and a response code related to
    redirecting.
    """
    REDIRECT_CODES = ['201', '301', '302']

    def redirect_location_for(self, result):
        if (result.code in self.REDIRECT_CODES):
            return (result.headers.get("Location") or
                    result.headers.get("location"))
        return None

    def execute(self, chain, request, env={}):
        result = chain.follow(request, env)
        location = self.redirect_location_for(result)
        if location:
            return self.redirect(location,
                                 request.headers.get("Content-Type"))

        return result

    def redirect(self, location, request_type):
        return restfulie.Restfulie.at(location).as_(request_type).get()



from dsl import Dsl


class Restfulie(object):
    """
    Restfulie DSL entry point.
    """

    @staticmethod
    def at(uri):
        """
        Create a new entry point for executing requests to the given uri.
        """
        return Dsl(uri)
